/* $Id: setup.c,v 1.9 2000/07/09 13:29:41 pavel Exp $
 * Setup the interrupt stuff.
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 1999 Harald Koerfgen
 */
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/mc146818rtc.h>
#include <linux/param.h>
#include <linux/init.h>
#include <linux/console.h>
#include <linux/ide.h>
#include <asm/mipsregs.h>
#include <asm/bootinfo.h>
#include <asm/irq.h>
#include <asm/reboot.h>
#include <asm/r39xx.h>
#include <asm/wbflush.h>
#include <asm/addrspace.h>

extern void add_wired_entry(unsigned long, unsigned long, 
 			    unsigned long, unsigned long);

extern asmlinkage void r39xx_handle_int(void);

extern void r39xx_machine_restart(char *command);
extern void r39xx_machine_halt(void);
extern void r39xx_machine_power_off(void);

extern struct ide_ops r39xx_ide_ops;
extern struct rtc_ops r39xx_rtc_ops;

extern int setup_r39xx_irq(int, struct irqaction *);

void (*board_time_init) (struct irqaction * irq);
void (*__wbflush) (void);

#undef PRINT_CONFIG

#ifdef PRINT_CONFIG
void debug_registers(void)
{
    /*
     * Debugging aid: print some configuration registers 
     */
    printk("MemConfig0:    0x%08x\n", MemConfig0);
    printk("MemConfig1:    0x%08x\n", MemConfig1);
    printk("MemConfig2:    0x%08x\n", MemConfig2);
    printk("MemConfig3:    0x%08x\n", MemConfig3);
    printk("MemConfig4:    0x%08x\n", MemConfig4);
    printk("\n");
    printk("ClockControl:  0x%08x\n", ClockControl);
    printk("\n");
    printk("IOcontrol:     0x%08x\n", IOControl);
    printk("MFIOOutput:    0x%08x\n", MFIOOutput);
    printk("MFIODirection: 0x%08x\n", MFIODirection);
    printk("MFIOInput:     0x%08x\n", MFIOInput);
    printk("MFIOSelect:    0x%08x\n", MFIOSelect);
    printk("IOPowerDown:   0x%08x\n", IOPowerDown);
    printk("MFIOPowerDown: 0x%08x\n", MFIOPowerDown);
    printk("\n");
    printk("IntEnable1:    0x%08x\n", IntEnable1);
    printk("IntEnable2:    0x%08x\n", IntEnable2);
    printk("IntEnable3:    0x%08x\n", IntEnable3);
    printk("IntEnable4:    0x%08x\n", IntEnable4);
    printk("IntEnable5:    0x%08x\n", IntEnable5);
    printk("IntEnable6:    0x%08x\n", IntEnable6);
    printk("\n");
    printk("IntStatus1:    0x%08x\n", IntStatus1);
    printk("IntStatus2:    0x%08x\n", IntStatus2);
    printk("IntStatus3:    0x%08x\n", IntStatus3);
    printk("IntStatus4:    0x%08x\n", IntStatus4);
    printk("IntStatus5:    0x%08x\n", IntStatus5);
    printk("IntStatus6:    0x%08x\n", IntStatus6);
    printk("\n");
    printk("CP0_STATUS:    0x%08x\n",  read_32bit_cp0_register(CP0_STATUS));
    printk("CP0_CAUSE:     0x%08x\n",  read_32bit_cp0_register(CP0_CAUSE));
    printk("CP0_CONF:      0x%08x\n",  read_32bit_cp0_register(CP0_CONF));
    printk("\n");
    printk("SPIData:	   0x%08x\n", SPIData);
    printk("SPIContol:	   0x%08x\n", SPIControl);
    printk("\n");
}
#endif


static void __init r39xx_irq_setup(void)
{

#ifdef PRINT_CONFIG
    debug_registers();
#endif

    /*
     * Turn all interrupts off
     */
    IntEnable1 = 0;
    IntEnable2 = 0;
    IntEnable3 = 0;
    IntEnable4 = 0;
    IntEnable5 = 0;
    IntEnable6 = 0;

    IntClear1 = 0xffffffff;
    IntClear2 = 0xffffffff;
    IntClear3 = 0xffffffff;
    IntClear4 = 0xffffffff;
    IntClear5 = 0xffffffff;
    IntClear6 = 0xffffffff;

    set_except_vector(0, r39xx_handle_int);

    set_cp0_status(ST0_BEV, 0);
    set_cp0_status(ST0_IM, read_32bit_cp0_register(CP0_STATUS) | IE_IRQ2 | IE_IRQ4);

    /* Set processor speed to 75MHz by zeroing reduced frequency
       bits in r39xx config register */
    set_cp0_conf(0xffff, (1 << CONFB_ICE) | (1 << CONFB_DCE)); 

}

/*
 * enable the periodic interrupts
 */
static __init void r39xx_time_init(struct irqaction *irq)
{
    /*
     * Here we go, enable periodic interrupts.
     */
    setup_r39xx_irq(0, irq);

    ClockControl |= CLK_ENTIMERCLK;
    RTCperiodTimer = PER_TIMER_COUNT;
    RTCtimerControl = TIM_ENPERTIMER;
    IntEnable5 |= INT5_PERIODICINT;
    ClockControl |= CLK_ENTIMERCLK;	/* Why is this done twice? */

#if defined(CONFIG_VTECH_HELIO)

    /* This is the wrong place for the following initialization code
       for the buttons and backlight.  It's just visiting
       here until the real drivers get finished. - jpick */

    /* Turn on Button Interrupts - to test int handler */
    MFIOSelect |=  MFIO_PIN_IO_KEY8 | MFIO_PIN_HOTSYNC_KEY;
    MFIODirection &= ~(MFIO_PIN_IO_KEY8 | MFIO_PIN_HOTSYNC_KEY);

    IntEnable3 = MFIO_PIN_IO_KEY8 | MFIO_PIN_HOTSYNC_KEY;
    IntEnable4 = MFIO_PIN_IO_KEY8 | MFIO_PIN_HOTSYNC_KEY;
    IntClear3 = MFIO_PIN_IO_KEY8 | MFIO_PIN_HOTSYNC_KEY;
    IntClear4 = MFIO_PIN_IO_KEY8 | MFIO_PIN_HOTSYNC_KEY; 

    IOControl |= IODEBSEL_MASK; /* Debounce all 7 inputs */
    IOControl &= ~IODIREC_MASK;  /* All seven are inputs */
    IOPowerDown = 0x7f;

    IntEnable5 |= (0x7f << INT5_POSINT_SHIFT) |
	    (0x7f << INT5_NEGINT_SHIFT);
    IntClear5 = (0x7f << INT5_POSINT_SHIFT) |
	    (0x7f << INT5_NEGINT_SHIFT);
    IntEnable6 |= INT6_IOPOSINT56 | INT6_IONEGINT56 | INT6_IOPOSNEGINT0;

    IntEnable5 |= INT5_POSONBUTINT | INT5_NEGONBUTINT;
    IntClear5 |= INT5_POSONBUTINT | INT5_NEGONBUTINT;
    PowerControl |= PWR_DBNCONBUTN;  /* Debounce On Button */

    /* Backlight Output */
    MFIOSelect |= MFIO_PIN_BACKLIGHT;
    MFIOOutput &= ~MFIO_PIN_BACKLIGHT;
    MFIODirection |= MFIO_PIN_BACKLIGHT; 



    /* Enable interrupts */
#ifdef CONFIG_VTECH_HELIO_EMULATOR
    IntEnable5 |= INT5_PERIODICINT;
    IntEnable6 |= INT6_GLOBALEN;
#else
    IntEnable6 |= INT6_GLOBALEN | INT6_PERIODICINT;
#endif

#else /* defined(CONFIG_VTECH_HELIO) */
    IntEnable6 |= INT6_GLOBALEN | INT6_PERIODICINT;
#endif

}

/*
 * Oh, well
 */
void r39xx_wbflush(void)
{
	/*
	 * The status of the writeback buffer is available
	 * via the Coprocessor 0 condition
	 */
    __asm__ __volatile__("
		.set	push
		.set	noreorder
		.set	mips2
		sync
		nop
	1:	bc0f	1b
		nop
		.set pop"
	);
}

void __init r39xx_setup(void)
{
    /*
     * This isn't strictly needed for Compaqs
     * but simplifies the keyboard driver a lot.
     */
    switch (mips_machtype) {
    case MACH_R39XX_SHARP_MOBILON:
	add_wired_entry (0x6c000f00, 0, KSEG3, 0);
	break;
    case MACH_R39XX_COMPAQ:
	add_wired_entry (0x10800f00, 0, KSEG3, 0);
	break;
    default:
        break;
    }

    irq_setup = r39xx_irq_setup;
    board_time_init = r39xx_time_init;

    mips_io_port_base = KSEG1ADDR(0x08000000);

    _machine_restart = r39xx_machine_restart;
    _machine_halt = r39xx_machine_halt;
    _machine_power_off = r39xx_machine_power_off;

#ifdef CONFIG_BLK_DEV_IDE
    ide_ops = &r39xx_ide_ops;
#endif

    rtc_ops = &r39xx_rtc_ops;

#ifdef CONFIG_FB
    conswitchp = &dummy_con;
#endif

#ifdef CONFIG_BLK_DEV_INITRD
    ROOT_DEV = MKDEV(RAMDISK_MAJOR, 0);
#endif

    __wbflush = r39xx_wbflush;

    MemConfig4 |= (MEM4_ENWATCH |
                   ((MEM4_WATCHTIME_VALUE & MEM4_WATCHTIMEVAL_MASK)
                    << MEM4_WATCHTIMEVAL_SHIFT));

}

