/* $Id: tpanel.c,v 1.7 2000/07/02 07:04:53 nop Exp $
 *
 * tpanel - Routines for Touch Panel on r39xx
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 2000 Jim Pick <jim@jimpick.com>
 *
 * Some parts taken from arch/mips/vr41xx/tpanel.c
 *
 */

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/poll.h>
#include <linux/miscdevice.h>
#include <linux/random.h>
#include <linux/init.h>
#include <asm/mipsregs.h>
#include <asm/system.h>
#include <asm/r39xx.h>
#include <asm/betty.h>
#include "sib.h"

static int pen_state = 0;
static int sample_counter;
static int sample_x, sample_y;

static wait_queue_head_t tpanel_wait;

#define BUFFSIZE 128

struct tpanel_sample {
	unsigned int state;
	unsigned int x;
	unsigned int y;
};

static struct tpanel_sample samplebuf[BUFFSIZE];
static struct tpanel_sample *samplebuf_head = samplebuf;
static struct tpanel_sample *samplebuf_tail = samplebuf;

struct timer_list tpanel_timer;

/* Prototypes */
static void tpanel_getxy(void);


static void tpanel_interrupt_mode(void)
{
	/* Clear All Interrupts */
	sib_write_check(BETTY_REG_INT_CLEAR_STATUS, 0);
	sib_write_check(BETTY_REG_INT_CLEAR_STATUS, 0xff);

	/* Start touch screen off in "interrupt mode" */

	sib_write_check(BETTY_REG_TOUCHSCREEN_CTRL,
			BETTY_TSMX_POW | BETTY_TSPX_POW |
			BETTY_TSMY_GND | BETTY_TSPY_GND |
			(0 << BETTY_TSC_MODE_SHIFT));

	sib_write_check(BETTY_REG_FALLING_INT_ENABLE, 0);
	sib_write_check(BETTY_REG_FALLING_INT_ENABLE, BETTY_TSPX_FAL_INT);

	sib_write_check(BETTY_REG_INT_CLEAR_STATUS, 0);
	sib_write_check(BETTY_REG_INT_CLEAR_STATUS, BETTY_TSPX_INT);

	/* Enable SIBIRQ pin interrupts (wired to interrupt on betty?) */
	IntEnable1 |= INT1_SIBIRQPOSINT;
	IntClear1 = INT1_SIBIRQPOSINT;
}


static void tpanel_readx_mode(void)
{
	/* Take touch screen out of "interrupt mode" */

	sib_write_check(BETTY_REG_FALLING_INT_ENABLE, 0);
	sib_write_check(BETTY_REG_INT_CLEAR_STATUS, 0);
	sib_write_check(BETTY_REG_INT_CLEAR_STATUS, BETTY_TSPX_INT);

		/* Get X Coordinate */

	sib_write_check(BETTY_REG_TOUCHSCREEN_CTRL,
			BETTY_TSPX_POW | BETTY_TSMX_GND |
			(2 << BETTY_TSC_MODE_SHIFT) |
			BETTY_TSC_BIAS_ENA);

	sib_write_check(BETTY_REG_ADC_CTRL,
			(BETTY_ADC_INPUT_TSPY <<
			 BETTY_ADC_INPUT_SHIFT) |
			BETTY_ADC_SYNC_ENA |
			BETTY_VREFBYP_CON |
			BETTY_ADC_ENA);

	sib_write_check(BETTY_REG_ADC_CTRL,
			(BETTY_ADC_INPUT_TSPY <<
			 BETTY_ADC_INPUT_SHIFT) |
			BETTY_ADC_SYNC_ENA |
			BETTY_VREFBYP_CON |
			BETTY_ADC_ENA |
			BETTY_ADC_START);
}


static void tpanel_ready_mode(void)
{
	sib_write_check(BETTY_REG_TOUCHSCREEN_CTRL,
			BETTY_TSMY_POW | BETTY_TSPY_GND |
			(2 << BETTY_TSC_MODE_SHIFT) |
			BETTY_TSC_BIAS_ENA);

	sib_write_check(BETTY_REG_ADC_CTRL,
			(BETTY_ADC_INPUT_TSPX <<
			 BETTY_ADC_INPUT_SHIFT) |
			BETTY_ADC_SYNC_ENA |
			BETTY_VREFBYP_CON |
			BETTY_ADC_ENA);

	sib_write_check(BETTY_REG_ADC_CTRL,
			(BETTY_ADC_INPUT_TSPX <<
			 BETTY_ADC_INPUT_SHIFT) |
			BETTY_ADC_SYNC_ENA |
			BETTY_VREFBYP_CON |
			BETTY_ADC_ENA |
			BETTY_ADC_START);
}

static void tpanel_timer_handler(unsigned long ptr)
{
/*	printk("T!\n"); */
	pen_state = 3;
	sib_read_check(BETTY_REG_TOUCHSCREEN_CTRL, tpanel_getxy);
}

static inline int tpanel_valid_data(int sample_x, int sample_y)
{
#if defined(CONFIG_VTECH_HELIO) && !defined(CONFIG_VTECH_HELIO_EMULATOR)
        /* To quote arch/mips/lib/memcpy.S: */
        /* Horror fix */

        /*  
         * Depending on the mood of my hardware, it can return bogus
         * low values (in particular, 4) for either X or Y samples
         * while the pen is moving.  I've never seen this on initial
         * pen hits.  I don't know whether this is a hardware problem
         * or a driver issue, but I need to throw them out.
         * 
         * Come to think of it, this is just the kind of symptom we'd
         * get if the sib driver was missing values.  Not surprising,
         * given the absolutely *wonderful* docs on how the SIB/tpanel
         * system works. :-( Will investigate later and back this 
         * out. ---nop@nop.com
         */

        if (sample_x < 10 || sample_y < 10) {
                return 0;
        }
#endif
        return 1;
}

static void tpanel_getxy(void)
{
	unsigned long value;
	int i;
	static int lastx, lasty;

/*	printk("%d!\n", pen_state); */
	switch( pen_state ) {

	case 0:
		/* Pen down interrupt recieved */

		tpanel_readx_mode();

		sample_counter = 0;
		lastx = -1;
		lasty = -1;

		pen_state = 1;
		sib_read_check(BETTY_REG_ADC_DATA, tpanel_getxy);
	     
		break;

	case 1:
		/* Some X Coordinate data was recieved from ADC */ 

		value = SIBSubFrame0Status;
		
		if (!(value & BETTY_ADC_DAT_VAL)) {
			if (++sample_counter < 60)
				sib_read_check(BETTY_REG_ADC_DATA,
					       tpanel_getxy);
			else {
				/* Put touch screen back into
				   "interrupt mode" */
				tpanel_interrupt_mode();
				pen_state = 0;
			}
			break;
		}

		/* If valid, then fall through */

/*
		printk("X: %d\n", 
		       (value & BETTY_ADC_DATA_MASK) >> BETTY_ADC_DATA_SHIFT);
*/

		sample_x = (value & BETTY_ADC_DATA_MASK) >>
			BETTY_ADC_DATA_SHIFT;

		/* Now, get Y Coordinate */

		tpanel_ready_mode();

		sample_counter = 0;

		pen_state = 2;
		sib_read_check(BETTY_REG_ADC_DATA, tpanel_getxy);
	     
		break;

	case 2:

		/* Some Y Coordinate data was recieved from ADC */ 

		value = SIBSubFrame0Status;
		
		if (! (value & BETTY_ADC_DAT_VAL)) {
			if (++sample_counter < 60)
				sib_read_check(BETTY_REG_ADC_DATA,
					       tpanel_getxy);
			else {
				/* Put touch screen back into
				   "interrupt mode" */
				tpanel_interrupt_mode();

				pen_state = 0;
			}
			break;
		}

		/* If valid, then fall through */

/*
		printk("Y: %d\n", 
		       (value & BETTY_ADC_DATA_MASK) >> BETTY_ADC_DATA_SHIFT);
*/

		sample_y = (value & BETTY_ADC_DATA_MASK) >>
			BETTY_ADC_DATA_SHIFT;

                if (tpanel_valid_data(sample_x, sample_y)) {
                        if ((lastx == -1) && (lasty == -1))
                                samplebuf_head->state = 1;
                        else {
                                samplebuf_head->state = 2;
                        }
                        samplebuf_head->x = sample_x;
                        samplebuf_head->y = sample_y;
                        
                        if ((sample_x >= lastx + 3) ||
                            (sample_x <= lastx - 3) ||
                            (sample_y >= lasty + 3) ||
                            (sample_y <= lasty - 3)) {
                                
/*			printk("K: %d %d %d\n", samplebuf_head->state, 
			       sample_x, sample_y); */

                                samplebuf_head++;
                                if (samplebuf_head == samplebuf + BUFFSIZE)
                                        samplebuf_head = samplebuf;
                                
                                if (samplebuf_head == samplebuf_tail) {
				/* Overrun */
                                        
                                        samplebuf_tail++;
                                        if (samplebuf_tail == samplebuf + BUFFSIZE)
                                                samplebuf_tail = samplebuf;
                                        
				/* printk("tpanel: Overrun.\n"); */
                                }
                                
                                lastx = sample_x;
                                lasty = sample_y;
                                
                                wake_up_interruptible(&tpanel_wait);
                                
                        }
                }
       		/* Put touch screen back into "interrupt mode" */
		tpanel_interrupt_mode();

		/* Set up a timer to take another sample after
		   an interval */

		sample_counter = 0;
		init_timer(&tpanel_timer);
		tpanel_timer.function = tpanel_timer_handler;
		tpanel_timer.expires = jiffies + HZ/20; /* 20x per sec */
		add_timer(&tpanel_timer);

		break;

	case 3:

		/* Read Touchscreen bits */

		value = SIBSubFrame0Status;
		if ((value & BETTY_TSPX_LOW) &&
		    (value & BETTY_TSMX_LOW)) {

			/* Pen up */

			samplebuf_head->state = 0;
			samplebuf_head->x = lastx;
			samplebuf_head->y = lasty;

/*			printk("K: %d %d %d\n", 0, lastx, lasty); */

			samplebuf_head++;
			if (samplebuf_head == samplebuf + BUFFSIZE)
				samplebuf_head = samplebuf;

			if (samplebuf_head == samplebuf_tail) {
				/* Overrun */

				samplebuf_tail++;
				if (samplebuf_tail == samplebuf + BUFFSIZE)
					samplebuf_tail = samplebuf;

				/* printk("tpanel: Overrun.\n"); */
			}

			wake_up_interruptible(&tpanel_wait);

			/* Put touch screen back into "interrupt mode" */
			tpanel_interrupt_mode();

			pen_state = 0;

		}
		else {
			/* Pen still down */
/*			printk("0x%08x\n", value ); */

			/* Take another X/Y measure */

			tpanel_readx_mode();

			sample_counter = 0;

			pen_state = 1;
			sib_read_check(BETTY_REG_ADC_DATA, tpanel_getxy);
		}

		break;

	}
}




static void tpanel_interrupt(int irq, void *dev_id,
					 struct pt_regs * regs)
{
	unsigned long value;
	int i;

	if (pen_state == 0)
		tpanel_getxy();
}


static int open_tpanel(struct inode * inode, struct file * file)
{
	samplebuf_tail = samplebuf_head;

	return 0;
}


/*
 *  Read touch panel data.
 *
 *  Data format:
 *  unsigned short state: 0 = pen up
 *                        1 = pen down
 *  unsigned short x:     raw data
 *  unsigned short y:     raw data
 *
 *  No calibration is done on the driver side, that is expected to be done on
 *  the user side.
 */

static ssize_t read_tpanel(struct file * file,
	char * buffer, size_t count, loff_t *ppos)
{
	unsigned int state, x, y;
	size_t retcnt = 0;

	if (count < 6)
		return -EINVAL;

		// We need to access the circular buffer atomic to
		// anything else that will read or write it

	cli();

	while (samplebuf_head == samplebuf_tail) {
		if (file->f_flags & O_NONBLOCK) {
			sti();
			return -EAGAIN;
		}
		interruptible_sleep_on(&tpanel_wait);
		// sleep_on will sti for us
		if (signal_pending(current))
			return -ERESTARTSYS;
		cli();
	}
	do {
		state = samplebuf_tail->state;
		x = samplebuf_tail->x;
		y = samplebuf_tail->y;

		samplebuf_tail++;
		if (samplebuf_tail == samplebuf + BUFFSIZE)
			samplebuf_tail = samplebuf;

		sti();

		if (!access_ok(VERIFY_WRITE, buffer, 6))
			return -EFAULT;

		__put_user(state, (short*)(buffer + retcnt));
		__put_user(x, (short*)(buffer + retcnt + 2));
		__put_user(y, (short*)(buffer + retcnt + 4));

		retcnt += 6;
		if (retcnt > count - 6)
			return retcnt;

		cli();
	} while (samplebuf_head != samplebuf_tail);

	sti();

	return retcnt;
}

static unsigned int poll_tpanel(struct file *file, poll_table * wait)
{
	int comp;

	poll_wait(file, &tpanel_wait, wait);
	cli();
	comp = (samplebuf_head != samplebuf_tail);
	sti();
	if (comp)
		return POLLIN | POLLRDNORM;
	return 0;
}


struct file_operations r39xx_tpanel_fops = {
        read:           read_tpanel,
        poll:           poll_tpanel,
        open:           open_tpanel,
};

static struct miscdevice r39xx_tpanel = {
	R39XX_TPANEL_MINOR, "r39xxtpanel", &r39xx_tpanel_fops
};

void r39xx_tpanel_init(void)
{
	int i;

	init_waitqueue_head(&tpanel_wait);

	cli();

	/* SIB (for touchscreen, sound and modem) */
	sib_init();

	/* Set panel to "interrupt mode" */
	if ((i = request_irq(41, tpanel_interrupt, SA_INTERRUPT,
			"touchpanel", NULL))) {
		printk(KERN_ERR "tpanel: Cannot allocate irq. (%d)\n", i);
	}
	tpanel_interrupt_mode();

	sti();

	misc_register(&r39xx_tpanel);

}

__initcall(r39xx_tpanel_init);
