/*
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * SNI specific PCI support for NEC VR4122.
 *
 * Copyright (C) 2000 Mike McDonald
 */
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <linux/types.h>
#include <asm/byteorder.h>
#include <asm/io.h>
#include <asm/vr4122.h>

#ifdef CONFIG_PCI

#define PCI_MEMBASE     0x10000000
#define PCI_MEMSIZE     0x05E00000
#define PCI_IOBASE      0x16000000
#define PCI_IOSIZE      0x02000000
#define PCI_IOEXCL      0x00010000      // Exclude first 64KB
#define PCI_MAW1        0x100F9010
#define PCI_MAW2        0x140FD014
#define PCI_IOAW        0x160FD000

#define mkaddr(dev, reg)				\
do {							\
    if (dev->bus->number == 0) {			\
	       if (PCI_SLOT(dev->devfn) < 11 ||		\
		   PCI_SLOT(dev->devfn) >= 32 ||	\
		   PCI_FUNC(dev->devfn) >= 8 ||		\
		   reg >= 256)				\
		   return 0xFFFFFFFF;			\
	*(volatile u32 *)VR41XX_PCICONFAREG =		\
            (reg & 0xFC) |				\
	    ( PCI_FUNC(dev->devfn) << 8) |		\
	    (1 << PCI_SLOT(dev->devfn));		\
    } else {						\
	       if (dev->bus->number >= 256 ||		\
		   PCI_SLOT(dev->devfn) >= 32 ||	\
		   PCI_FUNC(dev->devfn) >=8 ||		\
		   reg >= 256)				\
		   return 0xFFFFFFFF;			\
	*(volatile u32 *)VR41XX_PCICONFAREG = 1 |	\
	    (reg & 0xFC) |				\
	    (dev->devfn << 8) |				\
	    (dev->bus->number << 16);			\
    }							\
} while (0);

/*
 * We can't address 8 and 16 bit words directly.  Instead we have to
 * read/write a 32bit word and mask/modify the data we actually want.
 */
static int pcimt_read_config_byte (struct pci_dev *dev,
                                   int where, unsigned char *val)
{
	u32 res;

	mkaddr(dev, where);
	res = *(volatile u32 *)VR41XX_PCICONFDREG;
	res = (le32_to_cpu(res) >> ((where & 3) << 3)) & 0xff;
	*val = res;

	return PCIBIOS_SUCCESSFUL;
}

static int pcimt_read_config_word (struct pci_dev *dev,
                                   int where, unsigned short *val)
{
	u32 res;

	if (where & 1)
		return PCIBIOS_BAD_REGISTER_NUMBER;
	mkaddr(dev, where);
	res = *(volatile u32 *)VR41XX_PCICONFDREG;
	res = (le32_to_cpu(res) >> ((where & 3) << 3)) & 0xffff;
	*val = res;

	return PCIBIOS_SUCCESSFUL;
}

static int pcimt_read_config_dword (struct pci_dev *dev,
                                    int where, unsigned int *val)
{
	u32 res;

	if (where & 3)
		return PCIBIOS_BAD_REGISTER_NUMBER;
	mkaddr(dev, where);
	res = *(volatile u32 *)VR41XX_PCICONFDREG;
	res = le32_to_cpu(res);
	*val = res;

	return PCIBIOS_SUCCESSFUL;
}

static int pcimt_write_config_byte (struct pci_dev *dev,
                                    int where, unsigned char val)
{
	mkaddr(dev, where);
	*(volatile u8 *)(VR41XX_PCICONFDREG + (where & 3)) = val;

	return PCIBIOS_SUCCESSFUL;
}

static int pcimt_write_config_word (struct pci_dev *dev,
                                    int where, unsigned short val)
{
	if (where & 1)
		return PCIBIOS_BAD_REGISTER_NUMBER;
	mkaddr(dev, where);
	*(volatile u16 *)(VR41XX_PCICONFDREG + (where & 3)) = le16_to_cpu(val);

	return PCIBIOS_SUCCESSFUL;
}

static int pcimt_write_config_dword (struct pci_dev *dev,
                                     int where, unsigned int val)
{
	if (where & 3)
		return PCIBIOS_BAD_REGISTER_NUMBER;
	mkaddr(dev, where);
	*(volatile u32 *)VR41XX_PCICONFDREG = le32_to_cpu(val);

	return PCIBIOS_SUCCESSFUL;
}

struct pci_ops vr4122_pci_ops = {
        pcimt_read_config_byte,
        pcimt_read_config_word,
        pcimt_read_config_dword,
        pcimt_write_config_byte,
        pcimt_write_config_word,
        pcimt_write_config_dword
};

void __init pcibios_update_irq(struct pci_dev *dev, int irq)
{
    int slot_num;
    unsigned char pin;

    slot_num = PCI_SLOT(dev->devfn);
    pci_read_config_byte(dev, PCI_INTERRUPT_PIN, &pin);
//     printk("fix IRQ for slot %d dev->irq = %d pin = %d, irq = %d\n",
// 	   slot_num, dev->irq, pin, irq);
    dev->irq = 45;
}

static void __init pcibios_fixup_irqs(void)
{
    struct pci_dev *dev;
    int slot_num;
    int cnt = 0;

    pci_for_each_dev(dev) {
        slot_num = PCI_SLOT(dev->devfn);
	pcibios_update_irq(dev, cnt++);
    }
}

void __init pcibios_fixup_bus(struct pci_bus *b)
{
    printk("PCI:fixup bus %d - %s\n", b->number, b->name);
}

void pcibios_update_resource(struct pci_dev *dev, struct resource *root,
                        struct resource *res, int resource)
{
}

static void pci_intr_handler(int irq, void *dev_id, struct pt_regs *regs)
{
    int mask = (1 << (VR41XX_IRQ_PCI - VR41XX_IRQ_GPIO0));
    
    if (irq == VR41XX_IRQ_PCI) {
	*VR41XX_GIUINTSTATL = mask;
// 	printk("PCI Interrupt %d by %s, sdb = 0x%04x, pci = 0x%04x\n", irq, (char *) dev_id, *NEC_HARRIER_SDBINTREG, *NEC_HARRIER_PCIINTREG);
// 	*NEC_HARRIER_RSTREG = 0x0;
    }
}

struct resource pci_mem =
{ "VR4122 PCI MEM", PCI_MEMBASE, PCI_MEMBASE+PCI_MEMSIZE-1 };
struct resource pci_io  =
{ "VR4122 PCI I/O", PCI_IOBASE, PCI_IOBASE+PCI_IOSIZE-1 };

void __init pcibios_init(void)
{
        struct pci_ops *ops = &vr4122_pci_ops;
	volatile int data;
	unsigned short mask = (1 << (VR41XX_IRQ_PCI - VR41XX_IRQ_GPIO0));
	unsigned int clkdiv;

	printk("Looking for NEC VR4122 PCI controller at 0x%x ...", VR41XX_PCIMMAW1REG);
	*VR41XX_PCIMMAW1REG   = PCI_MAW1;
	*VR41XX_PCIMMAW2REG   = PCI_MAW2;
	*VR41XX_PCIMIOAWREG   = PCI_IOAW;
	*VR41XX_PCITAW1REG    = 0x00081000;
	*VR41XX_PCIMBA1REG    = 0x0;
	*VR41XX_PCIMBA2REG    = 0x0;
	*VR41XX_PCITAW2REG    = 0x0;
	*VR41XX_PCIMAILBAREG  = 0x0;
	data = *VR41XX_BUSERRADREG;
	clkdiv = *NEC_HARRIER_PCICLKREG;
	*VR41XX_PCICLKSELREG  = 0;
	*VR41XX_PCITRDYVREG   = 0x64;
	*VR41XX_PCICACHELSREG = 0x8004;
	*VR41XX_PCIENREG      = 0x4;
	*VR41XX_PCICOMMABDREG = 0x7;
	*VR41XX_INTCNTSTAREG  = 0xFF00FF00;

 	request_resource(&iomem_resource, &pci_mem);
 	request_resource(&iomem_resource, &pci_io);

	/* enable PCI interrupts */
	*NEC_HARRIER_SDBINTMASK = 0xff;
	*NEC_HARRIER_PCIINTMASK = 0xff;
	*VR41XX_MPCIINTREG = 0x1;
	*VR41XX_GIUIOSELL &= ~mask;
// 	*VR41XX_GIUPIODL &= ~mask;
	*VR41XX_GIUINTTYPL &= ~mask;
	*VR41XX_GIUINTALSELL &= ~mask;
	*VR41XX_GIUINTHTSELL &= ~mask;
	*VR41XX_GIUINTSTATL = mask;
	*VR41XX_GIUINTENL |= mask;
    	request_irq(VR41XX_IRQ_PCI, pci_intr_handler, SA_SHIRQ, "PCI", "PCI");

	if (*VR41XX_PCIVENDORIDREG == PCI_VENDOR_ID_NEC && *VR41XX_PCIDEVICEIDREG == PCI_DEVICE_ID_NEC_VR4122)
	    printk("OK\n");
	else {
	    printk("NOT FOUND!!\n");
	    return;
	}

        pci_scan_bus(0, ops, NULL);
	pcibios_fixup_irqs();

// 	printk("\nVR41XX_IRQ_PCI bit mask = 0x%02x\n", mask);
// 	printk("VR41XX_SYSINT1REG  = 0x%04x\n", *VR41XX_SYSINT1REG);
// 	printk("VR41XX_MSYSINT1REG = 0x%04x\n", *VR41XX_MSYSINT1REG);
// 	printk("VR41XX_GIUINTLREG  = 0x%04x\n", *VR41XX_GIUINTLREG);
// 	printk("VR41XX_MGIUINTLREG = 0x%04x\n", *VR41XX_MGIUINTLREG);
// 	printk("VR41XX_MPCIINTREG  = 0x%04x\n", *VR41XX_MPCIINTREG);
// 	printk("VR41XX_GIUIOSELL   = 0x%04x\n", *VR41XX_GIUIOSELL);
// 	printk("VR41XX_GIUPIODL    = 0x%04x\n", *VR41XX_GIUPIODL);
// 	printk("VR41XX_GIUINTENL   = 0x%04x\n", *VR41XX_GIUINTENL);
// 	printk("VR41XX_GIUINTTYPL  = 0x%04x\n", *VR41XX_GIUINTTYPL);
// 	printk("VR41XX_GIUINTALSELL= 0x%04x\n", *VR41XX_GIUINTALSELL);
	
// 	NEC_HARRIER_DISPLAY_LEDS(0x1234);
}

int __init pcibios_enable_device(struct pci_dev *dev)
{
        /* Not needed, since we enable all devices at startup.  */
        return 0;
}

void __init pcibios_align_resource(void *data, struct resource *res, unsigned long size)
{
}

char * __init pcibios_setup(char *str)
{
        /* Nothing to do for now.  */

        return str;
}

#define CMUCLKMSK	0x40
#define CMURST		0x42
#define MSYSINT1REG	0x6C

static void __init pci_fixup_4173(struct pci_dev *d)
{
    unsigned long ioaddr;
    int a;

    ioaddr = pci_resource_start (d, 0);
    printk("PCI: initializing the NEC 4173 at I/O of 0x%x\n", ioaddr);
    pcimt_write_config_byte(d, 0x40, 1);
    outw(0x1000, ioaddr + CMUCLKMSK);
    for (a=0; a<190000; a++); 
    outw(0x1dff, ioaddr + CMUCLKMSK);
    for (a=0; a<190000; a++); 
    outw(0x0f,   ioaddr + CMURST);
    for (a=0; a<190000; a++); 
    outw(0x0,    ioaddr + CMURST);
    outw(0x01ff, ioaddr + MSYSINT1REG);

    pcimt_read_config_dword(d, PCI_BASE_ADDRESS_0, &a);
    a = a >> 16;
    NEC_HARRIER_DISPLAY_LEDS(a);
}

static void __init pci_fixup_lance(struct pci_dev *d)
{
    long ioaddr;

    ioaddr = pci_resource_start (d, 0);
    printk("PCI: initializing the AMD Lance at I/O of 0x%x\n", ioaddr);
}

struct pci_fixup pcibios_fixups[] = {
        { PCI_FIXUP_HEADER,     PCI_VENDOR_ID_NEC,    PCI_DEVICE_ID_NEC_VRC4173,    pci_fixup_4173 },
        { PCI_FIXUP_HEADER,     PCI_VENDOR_ID_AMD,    PCI_DEVICE_ID_AMD_LANCE,    pci_fixup_lance },
        { 0 }
};

void *pci_alloc_consistent(struct pci_dev *hwdev, size_t size,
                           dma_addr_t *dma_handle)
{
        void *ret;
        int gfp = GFP_ATOMIC;
        int order = get_order(size);

        if (hwdev == NULL || hwdev->dma_mask != 0xffffffff)
                gfp |= GFP_DMA;
        ret = (void *)__get_free_pages(gfp, order);

        if (ret != NULL) {
                memset(ret, 0, size);
                *dma_handle = virt_to_bus(ret);
        }
        dma_cache_wback_inv((int) ret, PAGE_SIZE << order);
        return KSEG1ADDR(ret);
}

void pci_free_consistent(struct pci_dev *hwdev, size_t size,
                         void *vaddr, dma_addr_t dma_handle)
{
        free_pages((unsigned long)vaddr, get_order(size));
}

#endif /* CONFIG_PCI */
