/* Linker script for xip rom.
 * Adapted from arch/mips/ld.script.little by Bradley D. LaRonde.
 * No warranty.  Not even a little one.
 */

OUTPUT_FORMAT("elf32-littlemips")
OUTPUT_ARCH(mips)
ENTRY(kernel_entry)
SECTIONS
{
  /* Put text sections in rom at location specified by LOADADDR in arch/mips/Makefile. */
  .text :
  {
    /* _ftext is needed by arch/mips/mm/init.c, but I think for xip rom this might have to change. */
    _ftext = . ;
    *(.text)
    *(.rodata)
    *(.rodata1)

    /* __init_begin and __init_end are needed by arch/mips/mm/init.c
     * only to report "init size".
     */
    __init_begin = . ;

    /* For xip rom, init text is just like regular text. */
    *(.text.init)

    /* These sections are used in init/main.c. */
    __setup_start = .;
    *(.setup.init)
    __setup_end = .;
    __initcall_start = .;
    *(.initcall.init)
    __initcall_end = .;

    __init_end = . ;

    /* .gnu.warning sections are handled specially by elf32.em.  */
    *(.gnu.warning)

  } =0

  /* I put this here, but I don't think the kernel uses .rodata. */
  .rodata : { *(.rodata) }

  . = ALIGN(16);		/* Exception table */
  __start___ex_table = .;
  __ex_table : { *(__ex_table) }
  __stop___ex_table = .;

  __start___ksymtab = .;	/* Kernel symbol table */
  __ksymtab : { *(__ksymtab) }
  __stop___ksymtab = .;

  _etext = .;

  /* Put data sections in ram, but specify load address in rom. */
  .data : AT( _etext /* ADDR(.text) + SIZEOF(.text) */ )
  {
    _fdata = . ;
    *(.data)

    /* For xip rom, init data is same as regular data. */
    *(.data.init)

    /* init_task_union must be aligned to 8kb */
    . = ALIGN(8192);
    *(.data.init_task_union)

    . = ALIGN(4096);
    *(.data.idt)

    . = ALIGN(32);
    *(.data.cacheline_aligned)

    CONSTRUCTORS
    _edata  =  .;
  }

  .bss       :
  {
    *(.dynbss)
    *(.bss)
    *(.sbss)
    *(.scommon)
    *(COMMON)
    _end = . ;
  }

  /* Discard these sections */
  /DISCARD/ : {
	*(.text.exit)
	*(.data.exit)
	*(.exitcall.exit)
  }

  /* Stabs debugging sections.  */
  .stab 0 : { *(.stab) }
  .stabstr 0 : { *(.stabstr) }
  .stab.excl 0 : { *(.stab.excl) }
  .stab.exclstr 0 : { *(.stab.exclstr) }
  .stab.index 0 : { *(.stab.index) }
  .stab.indexstr 0 : { *(.stab.indexstr) }
  .comment 0 : { *(.comment) }
}
